function [ S, I, P, D, WS, WI ] = comp_inertia_3d( m1, m2 )
% Computes the tensor of intertia using the mass contained between m1 and m2
% Input:
%   m1      : outer mesh
%   m2      : inner mesh, must be of identical connectivity and orientation as m1 (optional, if no second mesh, use [])
%   method  : choses the method: 1 = divergence theorem
% Output:
%   S       : vector of all 10 moments of a rigid body
%   I       : full tensor of intertia
%   P       : principal axes of inertia
%   D       : magnitude of pricipal moments (eigenvalues)
%   SS      : moments of inertia with respect to world-coordinate frame
%   II      : tensor of intertia with respect to world-coordinate frame (if object is
%             rotated around Y-axes, which is assumed to be the rotation axis. 
%
% -------------------------------------------------------------------------
% copyright :   (c) 2014 Przemyslaw Musialski
% email     :   pm@cg.tuwien.ac.at
% created   :   22/11/2014
% edited    :   22/11/2014
% -------------------------------------------------------------------------

if ~exist('m1', 'var')
    disp('No input mesh. Exit.')
    return
end

% -------------------------------------------------------------------------
if ~exist('m2', 'var') || isempty(m2)
        
    % divergence theorem method [Eberly 2003]
    % get counters
    nf   = length(m1.f);
    nv   = length(m1.v);
    
    % get points in 3-block (for each triangle)
    p1   = m1.v(m1.f,:);
    n    = m1.n;
    
    % reorder points, each row is a triangle, each column is a coordinate (fnx9)
    t1   = p1(1:nf,:);
    t2   = p1(nf+1:2*nf,:);
    t3   = p1(2*nf+1:3*nf,:);
        
    % [Eberly 2003]
    S = comp_moments_3d(t1,t2,t3);
    %[S, ~, ~] = comp_moments_grad_3d(t1,t2,t3,m1.f,n);
    
    % -------------------------------------------------------------------------
else
    
    % num vertices
    np = length(m1.v);        
    % triangle index for double-mesh (flip the second half, since outer vertices are first)
    tidx = [ m1.f; [np+m2.f(:,2) , np+m2.f(:,1) , np+m2.f(:,3)] ];            
    % number of t-indices
    nf2   = length(tidx);     
    % get all points
    p = [ m1.v ; m2.v ];     
    % get points in 3-block (for each triangle)
    p = p(tidx,:);
    % compute moments
    [ S ] = comp_moments_3d( p(1:nf2,:), p(nf2+1:2*nf2,:), p(2*nf2+1:3*nf2,:) );
end;


if nargout>1
    I = [
        S(5),  S(8),  S(10);
        S(8),  S(6),  S(9);
        S(10), S(9),  S(7);
        ];
    
    if nargout > 2
        [P,D] = eig(I);
        
        d = diag(D);
        [d,idx] = sort(d,'descend');
        D = D(:, idx);
        P = P(:, idx);                     
        P(:,1) = d(1)*P(:,1);
        P(:,2) = d(2)*P(:,2);
        P(:,3) = d(3)*P(:,3);
                
        Ia = 0.5*(S(5) + S(7) - sqrt(4*S(10)^2 + S(5)^2 - 2*S(5)*S(7) + S(7)^2));
        Ib = 0.5*(S(5) + S(7) + sqrt(4*S(10)^2 + S(5)^2 - 2*S(5)*S(7) + S(7)^2)); 
        
        WS = S;
        WS(5) = Ia;
        WS(7) = Ib;
        WS(10) = 0;
        
        WI = I;        
        WI(1,1) = Ia;
        WI(3,3) = Ib;
        WI(1,3) = 0;
        WI(3,1) = 0;
    end
end

end
