function [ L, G, D ] = comp_laplacian3d_eig( m1, k, type, use_full, weights )
% computes the Laplacian and its k eigen-bases
% of the given mesh m
% input:
%   m1      : the input mesh
%   k       : number of eigen-bases
%   type    : type of the Laplacian
%
% output:
%   L: Laplacian for the n-polygon
%   G: nxk matrix containing the k eigen-vectors (eigen-bases)
%   D: the corresponding eigenvectors (nxn)
%
% -------------------------------------------------------------------------
% copyright :   (c) 2014-2015 Przemyslaw Musialski
% email     :   pm@cg.tuwien.ac.at
% created   :   10/01/2015
% edited    :   05/15/2015
% -------------------------------------------------------------------------

if ~exist('m1', 'var')
    disp('No input mesh. Exit.')
    return
end
if ~exist('use_full', 'var')
    use_full = 0;    
end
if ~exist('weights', 'var')
    weights = 'simple';    
end


%% alec jacobson version

if strcmp(type,'uniform')
    A = adjacency_matrix(m1.f);
    L = A - diag(sum(A));
else
    L = cotmatrix(m1.v, m1.f);        
    %L = L./M;
end

% compute k eigen-functions
if use_full 
    if strcmp(weights,'general')
        M = massmatrix(m1.v,m1.f, 'voronoi');
        [G,D] = eig(full(L),full(M));
        
    elseif strcmp(weights,'area')
        MD  = full(diag(massmatrix(m1.v,m1.f, 'voronoi')));
        Aij = 1./sqrt(MD*MD');
        Aij(logical(eye(size(Aij)))) = 0;  
        Lij = (Aij.*L);        
        Lij(logical(eye(size(Lij)))) = -sum(Lij,2);
        [G,D] = eig(full(Lij));
        
    else
        [G,D] = eig(full(0.5*L));
    end;
    d = diag(D);
    [~,I] = sort(d,'descend');    
    D = D(I, I);
    G = G(:, I);
    G = G(:,1:k);
    
else
    opts.v0 = full(ones(size(L(:,1))));
    opts.issym = 1;
    %[n-by-1 vector | {randomly generated by rand}]
    
    if strcmp(weights,'general')
        M = massmatrix(m1.v,m1.f, 'voronoi');        
        [G,D] = eigs(L, M, k, 'sm', opts);                
        
    elseif strcmp(weights,'area')
        MD  = full(diag(massmatrix(m1.v,m1.f, 'voronoi')));
        Aij = 1./sqrt(MD*MD');
        Aij(logical(eye(size(Aij)))) = 0;  
        Lij = (Aij.*L);        
        Lij(logical(eye(size(Lij)))) = -sum(Lij,2);        
        [G,D] = eigs(Lij, k, 'sm', opts);
    else
        [G,D] = eigs(0.5*L, k, 'sm', opts);
    end
end


end

