function [ d, dn, m2 ] = comp_max_offset_3d( m1, type, iter, threshold, rate, w0, const_off )
% Compute the max allowed offset
% Input is the mesh and the type
%   Type:
%     'contraction' :   use mesh contaction, this method also delivers trajectories
%   Iter:
%       number of iterations for contraction
%
% Output:
%   d  : max displacement
%   dn : normals
%   m2 : max-offset surface
%
% -------------------------------------------------------------------------
% copyright :   (c) 2014-2015 Przemyslaw Musialski
% email     :   pm@cg.tuwien.ac.at
% created   :   11/22/2014
% edited    :   04/26/2015
% -------------------------------------------------------------------------

if(~exist('type','var'))
    type = 'delaunay';
end
if(~exist('iter','var'))
    iter = 20;
end
if(~exist('threshold','var'))
    threshold = 1e-2;
end
if(~exist('rate','var'))
    rate = 2.0;
end
if(~exist('w0','var'))
    w0 = 1.0;
end
if(~exist('const_off','var'))
    const_off = 0.0;
end

if strcmp(type,'contraction')
    %% compute max offset by difference to contrated mesh
    % this contraction algortihm is based on the paper:
    % Au, O.K.-C., Tai, C.-L., Chu, H.-K., Cohen-Or, D., and Lee, T.-Y. 2008. 
    % Skeleton extraction by mesh contraction. ACM Transactions on Graphics 27, 3, 1.
    
    %area_mode = 'onering';
    area_mode = 'full';
    
    % apply const offset
    % off const_off=0 nothing happens
    v0   = m1.v;
    m1.v = m1.v - const_off*m1.n;
    
    L0 = cotmatrix(m1.v, m1.f, 0);
    A0 = full(sum(massmatrix(m1.v, m1.f, area_mode),2));
    
    % iteratively contract mesh
    disp('---- Contract mesh ...');
    
    mi = m1;
    
    n  = length(m1.v);
    E  = 1e-7*speye(n);
    Li = L0;
    W0 = w0*speye(n);
    WH = 1.0*W0;
    
    WL = 1.0*speye(n) * 1e-3 * sqrt(sum(sum(A0))/n);   
    sl = rate;
    z = zeros(n,3);
    V0 = comp_volume_3d( m1 );
    v = mi.v;
    
    for i=1:iter
        
        % solve
        b = WH*v;
        A = WH - WL*Li;       
        v = A\b;
        
        % compute volume and break on thresh
        mi.v = real(v);
        Vi = comp_volume_3d(mi);
        if (Vi/V0 < threshold)
            break;
        end
        
        % compute vertex reagion areas
        Ai = full(sum(massmatrix(mi.v,mi.f, area_mode),2));
        
        % prevent divison by 0
        Ai(abs(Ai)<1e-12) = 1e-12;        
        Ar = A0./Ai;
        
        % update arraction contrtaints
        Ar = sqrt(Ar);
        Ar = real(Ar);
        WH = w0 * spdiags(Ar,0,n,n);
        
        % compute cotan Laplacian
        Li = cotmatrix(v, mi.f, 0);
        WL = sl*WL;       
    end
    
    dn = mi.v-v0;
    d = sqrt(sum(dn.*dn,2));
    dn = bsxfun(@rdivide,dn,d);
    m2 = mi;
    
    if 0
        disp('---- Plot mesh ...');
        clf;
        set(gcf, 'renderer', 'opengl');
        set(gcf,'DefaultLineLineSmoothing','on');
        
        m2 = mesh;
        m2.v = m2.v+1.0*dn;
        plotMesh( m2, 'efbl', [1,0,0], [0.5,0,0.5] ); hold on;
        plotMesh( mesh, 'fbl', [0.81,0.81,0.81] ); hold on;
        %axis vis3d;
        
        % plot displacement
        if 0
            temp = [mi.v'; mesh.v'; nan(3,size(mi.v,1))];
            temp = reshape(temp,3,[]);
            line(temp(1,:),temp(2,: ), temp(3,: ),'color', [0.2, 0.8, 0.4]);
        end
        
    end
    return;    
else
    error(['' type ' is not a supported method']);
    
end

end

