 // ------------------------------------------------------------------------ -
 // copyright : (c)2014 - 2015 Przemyslaw Musialski
 // email     : pm@cg.tuwien.ac.at
 // created   : 11 / 16 / 2014
 // edited	  : 04 / 26 / 2015
 // this implements the pseudocode from [Eberly 2003]:
 //------------------------------------------------------------------------ -

#include <math.h>
#include <matrix.h>
#include <mex.h>


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    /* pointers to input & output matrices*/            
    double *V1 = mxGetPr(prhs[0]); /* 1 input matrix */
    double *V2 = mxGetPr(prhs[1]); /* 2 input matrix */
    double *V3 = mxGetPr(prhs[2]); /* 3 input matrix */    
        
    /* create output matrix M */
    plhs[0] = mxCreateDoubleMatrix(10, 1, mxREAL);        
    double *M = mxGetPr(plhs[0]);
    
    /* check dimensions of input matrices */
    size_t mrows,mb,ncols,nb;       /* matrix dimensions */
    mrows = mxGetM(prhs[0]);
    ncols = mxGetN(prhs[0]);
    mb = mxGetM(prhs[1]);
    nb = mxGetN(prhs[1]);    
    if (mrows != mb || ncols != nb )
    {
        mexErrMsgIdAndTxt("MATLAB:matchdims",
                "Dimensions of matices do not match.");
    }

    const double oneDiv6 =   (1.0/6.0);
    const double oneDiv24 =  (1.0/24.0);
    const double oneDiv60 =  (1.0/60.0);
    const double oneDiv120 = (1.0/120.0);
    
    bool bodyCoords = true;    
    double mass, cmx, cmy, cmz;
    
    // order:  1, x, y, z, x^2, y^2, z^2, xy, yz, zx
    double intg[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        
    for (int j = 0; j < mrows; j++)
    {
        // get vertices of triangle i
        double x0 = V1[j + mrows * 0];
        double y0 = V1[j + mrows * 1];
        double z0 = V1[j + mrows * 2];
        
        double x1 = V2[j + mrows * 0];
        double y1 = V2[j + mrows * 1];
        double z1 = V2[j + mrows * 2];
        
        double x2 = V3[j + mrows * 0];
        double y2 = V3[j + mrows * 1];
        double z2 = V3[j + mrows * 2];

        // get edges and cross product of edges
        double a1 = x1-x0; 
        double b1 = y1-y0; 
        double c1 = z1-z0; 
        double a2 = x2-x0; 
        double b2 = y2-y0; 
        double c2 = z2-z0;
        double d0 = b1*c2-b2*c1; 
        double d1 = a2*c1-a1*c2; 
        double d2 = a1*b2-a2*b1;

        // compute integral terms in x
        double temp0, temp1, temp2;
        double f1x, f2x, f3x, g0x, g1x, g2x;        
        temp0   = x0 + x1;
        temp1   = x0 * x0;
        temp2   = temp1 + x1 * temp0;        
        f1x     = temp0 + x2;                 
        f2x     = temp2 + x2 * f1x; 
        f3x     = (x0 * temp1) + (x1 * temp2) + (x2 * f2x);
        g0x     = f2x + x0 * ( f1x + x0 ); 
        g1x     = f2x + x1 * ( f1x + x1 ); 
        g2x     = f2x + x2 * ( f1x + x2 );

        // compute integral terms in y
        double f1y, f2y, f3y, g0y, g1y, g2y;                
        temp0   = y0 + y1;
        temp1   = y0 * y0;
        temp2   = temp1 + y1 * temp0;        
        f1y     = temp0 + y2;                 
        f2y     = temp2 + y2 * f1y; 
        f3y     = (y0 * temp1) + (y1 * temp2) + (y2 * f2y);
        g0y     = f2y + y0 * ( f1y + y0 ); 
        g1y     = f2y + y1 * ( f1y + y1 ); 
        g2y     = f2y + y2 * ( f1y + y2 );
        
        // compute integral terms in z
        double f1z, f2z, f3z, g0z, g1z, g2z;                
        temp0   = z0 + z1;
        temp1   = z0 * z0;
        temp2   = temp1 + z1 * temp0;        
        f1z     = temp0 + z2;                 
        f2z     = temp2 + z2 * f1z; 
        f3z     = (z0 * temp1) + (z1 * temp2) + (z2 * f2z);
        g0z     = f2z + z0 * ( f1z + z0 ); 
        g1z     = f2z + z1 * ( f1z + z1 ); 
        g2z     = f2z + z2 * ( f1z + z2 );
        
        // update integrals
        intg[0] += d0*f1x;
        intg[1] += d0*f2x; 
        intg[2] += d1*f2y; 
        intg[3] += d2*f2z;
        intg[4] += d0*f3x; 
        intg[5] += d1*f3y; 
        intg[6] += d2*f3z;
        intg[7] += d0*(y0*g0x+y1*g1x+y2*g2x);
        intg[8] += d1*(z0*g0y+z1*g1y+z2*g2y);
        intg[9] += d2*(x0*g0z+x1*g1z+x2*g2z);
    }

    // appliy constants
    intg[0] *= oneDiv6;
    intg[1] *= oneDiv24;
    intg[2] *= oneDiv24;
    intg[3] *= oneDiv24;
    intg[4] *= oneDiv60;
    intg[5] *= oneDiv60;
    intg[6] *= oneDiv60;
    intg[7] *= oneDiv120;
    intg[8] *= oneDiv120;
    intg[9] *= oneDiv120;

    // mass
    M[0] = mass = intg[0];
    
    // center of mass
    M[1] = cmx = intg[1]/mass;
    M[2] = cmy = intg[2]/mass;
    M[3] = cmz = intg[3]/mass;
    
    if (bodyCoords)
    {     
        //double cmx2 = cmx*cmx;
        //double cmy2 = cmy*cmy;
        //double cmz2 = cmz*cmz;
        
        // inertia relative to center of mass
        M[4] =  intg[5] + intg[6] - mass*(cmy*cmy + cmz*cmz);
        M[5] =  intg[4] + intg[6] - mass*(cmz*cmz + cmx*cmx);
        M[6] =  intg[4] + intg[5] - mass*(cmx*cmx + cmy*cmy);
        M[7] = -(intg[7] - mass*cmx*cmy);
        M[8] = -(intg[8] - mass*cmy*cmz);
        M[9] = -(intg[9] - mass*cmz*cmx);
    }
    else
    {
        // inertia relative to world origin
        M[4] =  intg[5] + intg[6]; // Ixx
        M[5] =  intg[4] + intg[6]; // Iyy
        M[6] =  intg[4] + intg[5]; // Izz

        M[7] = -intg[7]; // Ixy
        M[8] = -intg[8]; // Iyz
        M[9] = -intg[9]; // Izx
    }
}