function [ h ] = plot_shape3d( shape, show_lines, marked )
% -------------------------------------------------------------------------
% renders a shape, which as defined as follows
%   filename: filename of the original mesh
%   m1      :   original, outer surface mesh
%   m2      :   inner surface mesh
% -------------------------------------------------------------------------
% additinal data which might be or might not be set:
%   S       :   vector of shape moments, which are:
%       v   :   volume between m1-m2
%       c   :   center of mass of the volume (com)
%       I   :   moments of intertia (xx,yy,zz,xy,yz,xz)
%   P       :   principal directions (scaled to their lenght)
%   p       :   p (roh) is the material density constant (p=1)
%   L       :   the Laplacian of m1
%   Ltype   :   type of the L
%   G       :   eigen-functions of L
%   D       :   eigen-values of L
%   k       :   number of used eigen-functions
%   d       :   displacement (offset) used to compute m2
%   dn      :   displacement normals (pointing in the displacemet dir)
%   dlb     :   displacement lower-bounds 
%   dub     :   displacement upper-bounds
%   A       :   area-matrix (mass matrix)
%   Atype   :   type of A
%   m3      :   skeleton surface (or max-offset surface)
%   stype   :   type of the skeleton
%   siter   :   number of iterations to compute Siter
% -------------------------------------------------------------------------
% additional information used to create the offset surface m2
%   solver  :   the used solver
%   display :   type of display
%   grad    :   own gradient used (default is 'off')
%   weight  :   weight used for the objective
%   
% -------------------------------------------------------------------------
% copyright :   (c) 2014 Przemyslaw Musialski
% email     :   pm@cg.tuwien.ac.at   
% created   :   22/11/2014
% edited    :   22/11/2014           
% -------------------------------------------------------------------------


if(~exist('show_lines','var'))
    show_lines = 0;
end

disp('---- Plot mesh ...');
clf;
hold on;
set(gcf, 'renderer', 'opengl');
set(gcf,'DefaultLineLineSmoothing','on');


if (isfield(shape, 'm1'))
    plotMesh( shape.m1, 'fbl', [1,0,0] );
    %plotMesh( shape.m1, 'fbl', [0.81,0.81,0.81] );
end
if (isfield(shape, 'm2'))
    plotMesh( shape.m2, 'efbl', [1,0,0], [0.5,0,0.5] );
end
if ~exist('marked', 'var')
    marked = [];
end

if 0 % (isfield(shape, 'S')) %draw principal-axes of inertia
    scatter3( shape.S(2), shape.S(3), shape.S(4), 100, 'blue', 'filled' );        
    ds = 1.0;
    c = shape.S(2:4);
    p = shape.P;
    draw_line3(c, c+(ds*p(:,1)), 'LineWidth', 0.01, 'ArrowDirection', 1, 'ArrowLength', 0.04, 'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'r', 'LineColor', 'r');
    draw_line3(c, c+(ds*p(:,2)), 'LineWidth', 0.01, 'ArrowDirection', 1, 'ArrowLength', 0.04, 'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'g', 'LineColor', 'g');
    draw_line3(c, c+(ds*p(:,3)), 'LineWidth', 0.01, 'ArrowDirection', 1, 'ArrowLength', 0.04, 'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'b', 'LineColor', 'b');
end

if 0 % draw axes
    draw_line3([0,0,0],  [0,1,0], 'LineWidth', 0.005, 'ArrowDirection', 1, 'ArrowLength', 0.02,  'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'black', 'LineColor', 'black');
    draw_line3([-1,0,0], [1,0,0], 'LineWidth', 0.002, 'ArrowDirection', 1, 'ArrowLength', 0.006, 'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'black', 'LineColor', 'black');
    draw_line3([0,0,-1], [0,0,1], 'LineWidth', 0.002, 'ArrowDirection', 1, 'ArrowLength', 0.006, 'ArrowIntend', 0.003, 'ArrowAngle', 30, 'ArrowColor', 'black', 'LineColor', 'black');
    scatter3( 0, 0, 0, 140, 'black', 'filled' );
end
camtarget([0,0,0]);
h = gcf;


    % plot displacement    
    if show_lines
        temp = [shape.m2.v'; shape.m1.v'; nan(3,size(shape.m1.v,1))];
        temp = reshape(temp,3,[]);
        line(temp(1,:),temp(2,: ), temp(3,: ),'color', [0.2, 0.8, 0.4]);
    end
    
    if 0
        %  show disp between original 
        temp = [shape.m1.v'; shape.m0.v'; nan(3,size(shape.m1.v,1))];
        temp = reshape(temp,3,[]);
        line(temp(1,:),temp(2,: ), temp(3,: ),'color', [0.2, 0.8, 0.4]);
    end
    
    % plot points with low displacement
    if 0
        %pp = shape.m1.v(shape.m2.d<0.0001,:); %#ok<UNRCH>
        pp = shape.m1.v(marked,:);
        scatter3( pp(:,1), pp(:,2), pp(:,3), 100, 'red', 'filled' );        
    end
    
    if 0 % plot some intersections
        % clamp outer offsets to y>=miny;
        dirs = bsxfun(@times, shape.doub, shape.m1.don);
        orig = shape.m1.v;
        [flag, u, v, t] = rayPlaneIntersection_vec (orig, dirs, [0,0,0], [1,0,0], [0,0,1]);
        %s.dolb(t<shape.doub & t>=0) = t(t<shape.doub & t>=0);
        pp = shape.m1.v( (t<shape.doub & t>=0) ,:); %#ok<UNRCH>
        scatter3( pp(:,1), pp(:,2), pp(:,3), 100, 'red', 'filled' );        
    end
    
    if 0 % plot centers of mass
        % com of the shell
        [M1,~,~, ~, ~, ~] = comp_inertia_3d( shape.m1, shape.m2 );
        com_shell = M1(2:4);
        scatter3( com_shell(1), com_shell(2), com_shell(3), 100, 'red', 'filled' );       
        % com of the solid
        [M1,~,~, ~, ~, ~] = comp_inertia_3d( shape.m1, [] );
        com1 = M1(2:4);
        scatter3( com1(1), com1(2), com1(3), 50, 'blue', 'filled' );       
        
        sqrt((com_shell-com1).^2)
    end

hold off;
%axis vis3d;
end

