
function [ din, dout, result, exitflag, output, time, fhist, fmap ] = solve_combined_offsets3d_fmincon( mesh, costfun, w, G, L, alpha0, k, bounds, solver, grad, rho, approach, display )
%   Computes the offset-surface along normals using eigen-functions
%   of the mesh. Uses non-linear solver.
%
%   Input:
%       mesh: input mesh
%       t   : target moments
%       w   : weights for the moments w1,w2,w3
%       G   : eigen-basis of the mesh
%       roh : mass coefficient
%       lb  : lower bounds for displacements
%       ub  : upper bounds for displacements
%       solver: type of the sovler (default: sqp)
%       grad: use gradient function (default: off)
%       display: how much info to display
%
%   Output:
%       d : vector of displacements per points
%       sa: total area of the surface between input and offset
%       result: vector of results for mass, com, inertia
%
% -------------------------------------------------------------------------
%   copyright :   (c) 2014-2015 Przemyslaw Musialski
%   email     :   pm@cg.tuwien.ac.at
%   created   :   01/10/2015
%   edited    :   10/19/2015
% -------------------------------------------------------------------------

% check args
if ~exist('solver','var')
    solver = 'sqp';
end
if ~exist('display','var')
    display = 'final';
end
if ~exist('grad','var')
    grad = 'off';
end
if ~exist('rho','var')
    rho = [1.0,1.0];
end

M1 = 1;
CX = 2; CY = 3; CZ = 4;
XX = 5; YY = 6; ZZ = 7;
XY = 8; YZ = 9; ZX = 10;


% get k single
% if alpha is double, the first half is the inner surface
% the second half ist the outer surface
%approach = (size(alpha0,1)==k);

% get counters
np   = length(mesh.v);
nf   = length(mesh.f);
zero0= zeros(size(alpha0,1),1);

% get bounds
lbi  = bounds(:,1);
ubi  = bounds(:,2);
lbo  = bounds(:,3);
ubo  = bounds(:,4);

% get the harmonic functions
G    = G(:,1:k);

% get points in 3-block (for each triangle)
% p0   = mesh.v;
% ni0  = mesh.din;
% no0  = mesh.don;
%p1   = p0(mesh.f,:);

% NEW add max offset to the surface itself
%ubi = mesh.ddiub;
%ni0 = mesh.ddivec;

% get base points for inner and outer offsets
pio0 = [ mesh.v ; mesh.v ];

% get normals for inner and outer offsets
nio0 = [ mesh.din ; mesh.don ];

% triangle index for double-mesh, flip the first half (inner surface)
tri_inout   = [ [mesh.f(:,2) , mesh.f(:,1) , mesh.f(:,3)]; np+mesh.f ];
tri_out     = [ mesh.f ];
tri_in      = [ mesh.f ];

% number of triangle-indices
ntri_inout   = length(tri_inout);
ntri_out     = length(tri_out);
ntri_in      = length(tri_in);

np_inout     = length(pio0);
np_out       = length(mesh.v);
np_in        = length(mesh.v);

    Cx_both = 0;
    Cz_both = 0;
    DCx2_both = [];
    DCz2_both = [];
    DP_shell = [];
    DP_out = [];
    P_shell=[];
    P_out=[];
    P_in=[];
    

max_offet    = 100;
% points with fixed values
if ~isempty(mesh.fixpointsi)
    pi_fix       = mesh.fixpointsi;        
    pi_fixval    = mesh.fixvaluesi;        
else
    pi_fix       = [];        
    pi_fixval    = [];        
end
if ~isempty(mesh.fixpointso)    
    po_fix       = mesh.fixpointso;        
    po_fixval    = mesh.fixvalueso;    
else    
    po_fix       = [];        
    po_fixval    = [];    
end
pio_fix      = [ pi_fix; np_in+po_fix ];
pio_fixval   = [ pi_fixval; po_fixval ]; 

if(approach==1) % only inside
     
    % numper points
    nn = np;
    
    % gradient 
    DdDa = G;
    DdDa(pi_fix,:) = 0;
    %G(pi_fix,:) = 0;
        
    % lower and upper bounds    
    lbi(pi_fix) = -max_offet;
    ubi(pi_fix) = +max_offet;
    bb   = [-lbi ; +ubi]; 
    BGG  = [-G ; +G];
    
elseif(approach==3) % only outside
    nn = np;                
    DdDa = G;
    DdDa(po_fix,:) = 0;        
    lbo(po_fix) = -max_offet;
    ubo(po_fix) = +max_offet;
    bb   = [-lbo ; +ubo];  
    BGG  = [-G ; +G];
    
else % both   
    nn = 2*np;
    
    DdDa = blkdiag(G,G); 
    DdDa(pio_fix,:) = 0;
    
    lbi(pi_fix) = -max_offet;
    ubi(pi_fix) = +max_offet;
    lbo(po_fix) = -max_offet;
    ubo(po_fix) = +max_offet;
    
    bb   = [-lbi; +ubi; -lbo; +ubo];
    BGG  = blkdiag([-G ; +G],[-G ; +G]);
        
end

fhist = [];
fmap = containers.Map;
fmap('Etot') = 0;
fmap('Ered') = 0;

P1 = [];
DmDd = []; ds1 = [];
M_out = []; P2 = [];
DM_out =[]; ds2 = []; 

% --------------------------
%options = optimoptions(@fmincon,...
options = optimset(...
    'Algorithm', solver,... % 'interior-point' , 'active-set' 'sqp'
    'GradObj',grad,...
    'GradConstr',grad,...
    'Display', display,...
    'MaxFunEvals', 1000*np,...
    'TolCon',1e-3,... % Tolerance on the constraint violation, a positive scalar. The default is 1e-6.
    'TolFun',1e-3,... % Termination tolerance on the function value, a positive scalar. The default is 1e-6.
    'TolX', 1e-5, ... % Relative changes in all elements of x is the normalized step vector. The default is 1e-10.
    'Hessian','bfgs',... % 'bfgs'{'lbfgs',positiveinteger}
    'MaxIter',1000);

if strcmp(costfun,'spin')
    
    %BGG  = [-G ];
    %bb   = [-lbi ];
    %options = optimset(options,'GradObj','on', 'GradConstr','on');
    tic;
    [alpha1,fval,exitflag,output,lambda,grad,hess] =      fmincon(@obj_spin, alpha0,  BGG,bb,  [],[],  [],[],  @const_spin, options );
    time = toc;
    obj_spin(alpha1);
    const_spin(alpha1);
        
elseif strcmp(costfun,'stand')
    
    tic;
    [alpha1,fval,exitflag,output,lambda,grad,hess] =      fmincon(@obj_stand_soft, alpha0,  BGG,bb,  [],[],  [],[],  @const_stand_soft, options );
    time = toc;
    obj_stand_soft(alpha1);
    const_stand_soft(alpha1);
    
elseif strcmp(costfun,'standup')
    
    tic;
    [alpha1,fval,exitflag,output,lambda,grad,hess] =      fmincon(@obj_standup, alpha0,  BGG,bb,  [],[],  [],[],  @const_standup, options );
    time = toc;
    obj_standup(alpha1);
    const_standup(alpha1);
    
elseif strcmp(costfun,'fish')
    
    tic;
    [alpha1,fval,exitflag,output,lambda,grad,hess] =      fmincon(@obj_fish, alpha0,  BGG,bb,  [],[],  [],[],  @const_fish, options );
    time = toc;
    obj_fish(alpha1);
    const_fish(alpha1);       
    
elseif strcmp(costfun,'bottle')
    
    %BGG  = [-G ];
    %bb   = [-lbi ];
    %options = optimset(options,'GradObj','off', 'GradConstr','off');
    tic;
    [alpha1,fval,exitflag,output,lambda,grad,hess] =      fmincon(@obj_bottle, alpha0,  BGG,bb,  [],[],  [],[],  @const_bottle, options );
    time = toc;
    obj_bottle(alpha1);
    const_bottle(alpha1);
    
else% strcmp(costfun,'default')
    tic;
    output = [];
    output.iterations = 0;
    exitflag = 0; 
    alpha1 = alpha0;
    obj3(alpha1);
    const3(alpha1);     
    time = toc;
end

if(approach==1)
    din  = G*alpha1;
    din(pi_fix) = pi_fixval; 
    dout = [];
elseif(approach==3)    
    din  = [];
    dout = G*alpha1;
    dout(po_fix) = po_fixval; 
else
    din  = G*alpha1(1:k);
    dout = G*alpha1(k+1:end);
    din(pi_fix) = pi_fixval;     
    dout(po_fix) = po_fixval; 
    %output = po_fix;
end;
exitflag %#ok<NOPRT>
result = P1; %[vs, cx,cy,cz,  ixx,iyy,izz, ixy,iyz,ixz];
% --- end of func


% =====================================================================

    function [f, gf] = obj3(a0)
        % this objective uses the divergence theorem to compute the moments
        
        % evaluate eigen-functions
        % di0    :   particular inner displacements per vertex
        % do0    :   outer displacements per vertex
        %di  = G*a0(1:k);
        %do  = G*a0(k+1:end);
        
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        [ P1, DmDd ] = comp_moments_div(d0);
        
        % objective
        [f, gf] = objective(a1);
        %[f, gf] = objective_spinit(a1);
        
    end
    function [ c, ceq, gc, gceq ] = const3(a0)
        
        % evaluate eigen-functions
        % di0    :   particular inner displacements per vertex
        % do0    :   outer displacements per vertex
        %         di  = G*a0(1:k);
        %         do  = G*a0(k+1:end);
        %         [ S, DmDd ] = comp_moments_div(di,do);
        
        [ c, ceq, gc, gceq ] = constraints();
        %[ c, ceq, gc, gceq ] = constraints_spinit();
        %[ c, ceq, gc, gceq ] = constraints_coca3();
        
        
    end
% ---------------------------------------------------------------------

% =====================================================================

    function [ f, gf ] = objective(a1)
        
        % this term minimizes the volume
        f1 = P1(M1)^1;
        % this term minimizes mass-times-hight of the center of mass (c(2)=cy)
        % the center of mass should be low, the mass as well
        f2 = P1(CY)^2;
        % this term minimizes the ratio of moments of inertia
        f3 = (P1(XX)/P1(YY))^1 + (P1(ZZ)/P1(YY))^1;
        %f3 = ixx + izz;
        % this term maximizes the spin-axis
        f4 = (1/P1(YY))^1;
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        % objective weighted by custom weights w
        f = w(1)*f1 + w(2)*f2 + w(3)*f3 + w(4)*f4 + w(6)*f5;
        %f = w(2)*(f1 * f2) + w(3)*f3 + w(4)*f4 + w(6)*f5;
        fhist = [fhist, f];
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
        
        if nargout>1
            % gradient
            ds1 = DmDd*DdDa;
            gf1 = 1*ds1(1, :)';
            gf2 = 2*P1(CY)*ds1(3, :)';
            
            dixx = ds1(5,:)';
            diyy = ds1(6,:)';
            dizz = ds1(7,:)';
            
            % if f3^2
            % -((2 ixx[x] (-iyy[x] (ixx^\[Prime])[x]+ixx[x] (iyy^\[Prime])[x]))/iyy[x]^3)
            %gf3x = -((2*ixx*(ixx*diyy-iyy*dixx))/(iyy*iyy*iyy));
            %gf3z = -((2*izz*(izz*diyy-iyy*dizz))/(iyy*iyy*iyy));
            %gf3 = gf3x+gf3z;
            
            % if f3^1
            % (iyy[x] Derivative[1][ixx][x] - ixx[x] Derivative[1][iyy][x])/iyy[x]^2
            gf3x = ((P1(YY)*dixx-P1(XX)*diyy)/(P1(YY)*P1(YY)));
            gf3z = ((P1(YY)*dizz-P1(ZZ)*diyy)/(P1(YY)*P1(YY)));
            gf3 = gf3x+gf3z;
            % simple
            %gf3 = dixx + dizz;
            
            % -((2 Derivative[1][iyy][x])/iyy[x]^3)
            %gf4 = -((2*diyy)/(iyy*iyy*iyy)); % from .^2
            % -(Derivative[1][iyy][x]/iyy[x]^2)
            gf4 = -((diyy)/(P1(YY)*P1(YY)));       % from .^1
            
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;
            
            gf = w(1)*gf1 + w(2)*gf2 + w(3)*gf3 + w(4)*gf4 + w(6)*gf5;
            %gf = w(2)*(f1*gf2 + f2*gf1) + w(3)*gf3 + w(4)*gf4 + w(6)*gf5;
        end
    end
    function [ c, ceq, gc, gceq ] = constraints()
        ex = 1;
        ceq = [
            w(2)*P1(2)^2; %cx
            w(2)*P1(4)^2; %cz
            w(5)*P1(8)^1; %ixy
            w(5)*P1(9)^1; %iyz
            w(5)*P1(10)^1 %izx
            ];
        c = [];
        
        if nargout>2
            % gradient
            ds1 = DmDd*DdDa;
            gceq = [
                w(2)*2*P1(2)*ds1(2,:);
                w(2)*2*P1(4)*ds1(4,:);
                w(5)*ds1(8,:);
                w(5)*ds1(9,:);
                w(5)*ds1(10,:)
                ]';
            gc = [];
        end
    end

% ---------------------------------------------------------------------

    function [ f, gf ] = obj_spin(a0)
        
        % get displacements
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        [ P1, DmDd ] = comp_moments_div(d0);
        %[ M, DmDd ] = comp_moments_volumes_div(d0);
        
             
        % this term minimizes the volume
        f1 = P1(M1)^1;
        
        % this term minimizes mass-times-hight of the center of mass (c(2)=cy)
        % the center of mass should be low, the mass as well
        %f2 = M(M1)*M(CY);
        f2 = P1(CY)^1;
 
        % this term minimizes the ratio of moments of inertia
        % numeric eigenvalues
        %[~,D] = eig([ M(5), M(10); M(10), M(7) ]);
        %f3 = (D(1,1)/iyy)^1 + (D(2,2)/iyy)^1;
        % analytic eigenvalues
        Ia = 0.5*(P1(XX) + P1(ZZ) - sqrt(4*P1(ZX)^2 + P1(XX)^2 - 2*P1(XX)*P1(ZZ) + P1(ZZ)^2));
        Ib = 0.5*(P1(XX) + P1(ZZ) + sqrt(4*P1(ZX)^2 + P1(XX)^2 - 2*P1(XX)*P1(ZZ) + P1(ZZ)^2));        
        %f3 = (Ia/M(YY))^1 + (Ib/M(YY))^1;
        f3 = (Ia/P1(YY))^2 + (Ib/P1(YY))^2;
        
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        %objective weighted by custom weights w
        %f = w(1)*f1 + w(2)*f2 + w(3)*f3 + w(6)*f5;
        f = w(2)*(f1 * f2) + w(3)*f3 + w(6)*f5;
        fhist = [fhist, f];
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
        
        if nargout>1
            % compute final derivatives
            ds1 = DmDd*DdDa;
            %DM_out = DmDd_out*DdDa;
            %DM_in = DmDd_in*DdDa;
            
            % gradient
            gf1 = ds1(M1, :)';
            %gf2 = M(M1)*DM(CY, :)' + M(CY)*DM(M1, :)';
            %gf2 = 2*M(CY)*DM(CY, :)';
            gf2 = ds1(CY, :)';
            
            DMXX = ds1(XX,:)';%DM5
            DMYY = ds1(YY,:)';%DM6            
            DMZZ = ds1(ZZ,:)';%DM7
            DMZX = ds1(ZX,:)';%DM10            
            % if f3^1 with eigs
            %(-(M5[x] + M7[x]) Derivative[1][M6][x] +  M6[x] (Derivative[1][M5][x] + Derivative[1][M7][x]))/M6[x]^2
            %gf3 = (-(M(XX) + M(ZZ))*DMYY +  M(YY)* (DMXX + DMZZ))/ M(YY)^2;
            % if f3^2 with eigs
            gf3 = (1/(P1(YY)^3))*2*(P1(XX)*P1(YY)*DMXX - P1(XX)^2*DMYY - 2*P1(ZX)^2*DMYY + 2*P1(YY)*P1(ZX)*DMZX + P1(ZZ)*(-P1(ZZ)*DMYY + P1(YY)*DMZZ));
                    
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;  
            
            %gf = w(1)*gf1 + w(2)*gf2 + w(3)*gf3 + w(6)*gf5;
            gf = w(2)*(f1*gf2 + f2*gf1) + w(3)*gf3 + w(6)*gf5;
        end
    end
    function [ c, ceq, gc, gceq ] = const_spin(a0)

        ceq = [
            w(2)*P1(CX)^1; %cx
            w(2)*P1(CZ)^1; %cz
            w(5)*P1(XY)^1; %ixy
            w(5)*P1(YZ)^1; %iyz
            ];
        c = [            
            w(2)*(-P1(CY));
            ];
        
        if nargout>2
            % gradient
            % is computed in the objective: 
            % DM = DmDd*DdDa;
            gceq = [
                w(2)*ds1(CX,:);
                w(2)*ds1(CZ,:);
                w(5)*ds1(XY,:);
                w(5)*ds1(YZ,:);
                %w(5)*Ds(10,:)
                ]';
            gc = [                
                w(2)*(-ds1(CY,:)); 
                ]';
        end
    end

% --------------------------------------------------------------------- 

    % This setup forces the COM to be over a region defined by a radius
    % if the radius is set to 0, the objective is like hard constraints
    % if the radius is Inf or the weight(5)==0, the function is soft
   
    function [ f, gf ] = obj_stand_soft(a0)
        
        % get displacements
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        [ P1, DmDd ] = comp_moments_div(d0);
             
        % this term minimizes the volume
        f1 = P1(M1)^1;
        
        % this term minimizes mass-times-hight of the center of mass (c(2)=cy)
        % the center of mass should be low, the mass as well
        f2 = P1(CY)^1 + P1(CX)^2 + P1(CZ)^2;
         
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        %objective weighted by custom weights w
        f = w(1)*f1 + w(2)*f2 +  w(6)*f5;
        %f = w(2)*(f1 * f2) + w(3)*f3 + w(4)*f4 + w(6)*f5;
        fhist = [fhist, f];
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
        
        if nargout>1
            % compute final derivatives
            ds1 = DmDd*DdDa;
            %DM_out = DmDd_out*DdDa;
            %DM_in = DmDd_in*DdDa;

            %gf1 = 2*M(M1)*DM(M1, :)';
            gf1 = ds1(M1, :)';
            gf2 = ds1(CY, :)' + 2*P1(CX)*ds1(CX, :)'+2*P1(CZ)*ds1(CZ, :)';
            
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;  
            
            gf = w(1)*gf1 + w(2)*gf2 + w(6)*gf5;
            %gf = w(2)*(f1*gf2 + f2*gf1) + w(3)*gf3 + w(6)*gf5;
        end
    end
    function [ c, ceq, gc, gceq ] = const_stand_soft(a0)

        ceq = [];
        c = [ 
            -P1(CY);
            w(5)*(P1(CX)^2 - mesh.R2);
            w(5)*(P1(CZ)^2 - mesh.R2);
            ];
        
        if nargout>2
            % gradient
            % is computed in the objective: 
            % DM = DmDd*DdDa;
            gceq = []';
            gc = [ 
                -ds1(CY,:);
                w(5)*(2*P1(CX)*ds1(CX, :));
                w(5)*(2*P1(CZ)*ds1(CZ, :));
                ]';
        end
    end

% ---------------------------------------------------------------------          

    function [ f, gf ] = obj_bottle(a0)
                
        % get displacements
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        [ P1, DmDd, ~,  ~, ~, ~,  P2, DmDd_in, ~ ] = comp_moments_volumes_div(d0);
        P1(M1) = rho(1)*P1(M1);
        P2(M1) = rho(2)*P2(M1);
                    
        % this term minimizes the shell volume
        f1 = P1(M1)^2;
        
        % this term minimizes the center of mass (c(2)=cy)        
        %f21 = M(CY)^2 + M(CX)^2 + M(CZ)^2;
        
        mm1 = P1(M1)+P2(M1);
        % optmize the joint COM of both 
        Cx_both = ((P1(M1)*P1(CX) + P2(M1)*P2(CX))/mm1);
        Cz_both = ((P1(M1)*P1(CZ) + P2(M1)*P2(CZ))/mm1);
        
        f2x = 0*P1(CX)^2 + (Cx_both)^2;
        f2y = 0*P1(CY)^1 + ((P1(M1)*P1(CY) + P2(M1)*P2(CY))/mm1)^1;
        f2z = 0*P1(CZ)^2 + (Cz_both)^2;
        f2  = f2x + f2y +f2z;
        
        % this term minimizes the center of mass (c(2)=cy)        
        %f22 = M_in(CY)^2 + M_in(CX)^2 + M_in(CZ)^2;
         
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        %objective weighted by custom weights w
        f = w(1)*f1 + w(2)*f2 + w(6)*f5;        
        fhist = [fhist, f];  
%        disp([Cx_both, P1(CX)]); 
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
                    
        if nargout>1                        
            ds1 = DmDd*DdDa; 
            ds2 = DmDd_in*DdDa(1:np_in, :);
            ds1(M1,:) = rho(1)*ds1(M1,:);
            ds2(M1,:) = rho(2)*ds2(M1,:);
            
            gf1  = (2*P1(M1)*ds1(M1, :))';
              
            % this is the derivative of Cx_both^2:
            DCx2_both = (1/((P1(M1) + P2(M1))^3))* 2*(P1(CX)*P1(M1) + ...
                P2(CX)*P2(M1))*(ds2(M1,:)*P1(M1)*(-P1(CX) + P2(CX)) + ...
                ds1(CX,:)*P1(M1)*(P1(M1) + P2(M1)) + ...
                P2(M1)*(ds1(M1,:)*(P1(CX) - P2(CX)) + ds2(CX,:)*(P1(M1) + P2(M1))));
            DCx2_both = DCx2_both';
            
            % this is the derivative of Cz_both^2:
            DCz2_both = (1/((P1(M1) + P2(M1))^3))* 2*(P1(CZ)*P1(M1) + ...
                P2(CZ)*P2(M1))*(ds2(M1,:)*P1(M1)*(-P1(CZ) + P2(CZ)) + ...
                ds1(CZ,:)*P1(M1)*(P1(M1) + P2(M1)) + ...
                P2(M1)*(ds1(M1,:)*(P1(CZ) - P2(CZ)) + ds2(CZ,:)*(P1(M1) + P2(M1))));
            DCz2_both = DCz2_both';
            
            Dcy1_both = (ds2(M1,:)*P1(M1)*(-P1(CY) + P2(CY)) + ds1(CY,:)*P1(M1)*(P1(M1) + P2(M1)) + ...
                P2(M1)*(ds1(M1,:)*(P1(CY) - P2(CY)) + ds2(CY,:)*(P1(M1) + P2(M1))))/(P1(M1) + P2(M1))^2;
            Dcy1_both = Dcy1_both';
            
            gf2 = DCx2_both+Dcy1_both+DCz2_both;
            
            %gf1 = 2*M(M1)*DM(M1, :)';
            %gf21 = 2*M(CY)*DM(CY, :)' + 2*M(CX)*DM(CX, :)'+2*M(CZ)*DM(CZ, :)';
            %gf22 = 2*M_in(CY)*DM_in(CY, :)' + 2*M_in(CX)*DM_in(CX, :)'+2*M_in(CZ)*DM_in(CZ, :)';
            
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;  
            
            gf = w(1)*gf1 + w(2)*gf2 + w(6)*gf5;
        end
    end
    function [ c, ceq, gc, gceq ] = const_bottle(a0)
               
        ceq = [];        
        c   = [
            w(5)*(-P1(CY));            
            Cx_both^2 - mesh.R2;
            Cz_both^2 - mesh.R2;
%             w(5)*(s1(CX)^2 - mesh.R2);
%             w(5)*(s1(CZ)^2 - mesh.R2);                        
%             w(5)*(s2(CX)^2 - mesh.R2);
%             w(5)*(s2(CZ)^2 - mesh.R2);
            ];
        
        %disp([Cx_both^2 - mesh.R2, Cx_both, P1(CX), P2(CX)]);  
                            
        if nargout>2    
            
            gceq = []';
            gc   = [
                w(5)*(-ds1(CY, :)),
                DCx2_both';
                DCz2_both';
%                 w(5)*(2*s1(CX)*ds1(CX, :));
%                 w(5)*(2*s1(CZ)*ds1(CZ, :));
%                 w(5)*(2*s2(CX)*ds1(CX, :));
%                 w(5)*(2*s2(CZ)*ds1(CZ, :));
                ]';
        end        
    end

% --------------------------------------------------------------------- 


    function [ f, gf ] = obj_fish(a0)
        
        % get displacements
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        %M, DmDd, p_inout,     M_out, DmDd_out, p_out,     M_in, DmDd_in, p_in
        [ P_shell,DmDd_shell,~,  P_out,DmDd_out,~,   P_in,~,~ ] = comp_moments_volumes_div(d0);
             
        % COM_shell = center of gravtiy
        % COM_out = center of boyoucy
        % center of masses
        f1x = (P_shell(CX)-P_out(CX))^2;  
        f1z = (P_shell(CZ)-P_out(CZ))^2;        
        % disY must be: Cy_out>Cy_shell
        
        % rroh = rho_water/rho_mat
        %rrho=rho(2)/rho(1); 
        f2 = (rho(2)*P_out(M1)-rho(1)*P_shell(M1))^2;
        
        % boyouncy point towards center
        f3 = P_out(CX)^2+P_out(CY)^2+P_out(CZ)^2;
        
        % maxmimize the distance between COG and POB
        f4 = P_shell(CY)-P_out(CY);
        
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        % objective
        f = w(1)*(f1x+f1z) + w(2)*f2 + w(3)*f3 + w(4)*f4 + w(6)*f5;                
        fhist = [fhist, f];
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
        P1 = P_shell;
        
        if nargout>1
            % compute final derivatives
            DP_shell = DmDd_shell*DdDa;
            DP_out = DmDd_out*DdDa(np_in+1:np_in+np_out,:);                        
            DP_shell(M1,:) = rho(1)*DP_shell(M1,:);
            DP_out(M1,:) = rho(2)*DP_out(M1,:);
                
            gf1x = 2*(P_out(CX) - P_shell(CX))*(DP_out(CX,:) - DP_shell(CX,:));
            gf1z = 2*(P_out(CZ) - P_shell(CZ))*(DP_out(CZ,:) - DP_shell(CZ,:));
            gf2  = 2*(rho(2)*P_out(M1) - rho(1)*P_shell(M1))*(rho(2)*DP_out(M1,:) - rho(1)*DP_shell(M1,:));
            
            gf3  = 2*(P_out(CX)*DP_out(CX,:)+P_out(CY)*DP_out(CY,:)+P_out(CZ)*DP_out(CZ,:));
            
            gf4  = DP_shell(CY,:)-DP_out(CY,:);
            
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;  
            
            gf = w(1)*(gf1x+gf1z)' + w(2)*gf2' + w(3)*gf3' + w(4)*gf4' + w(6)*gf5;                        
        end
    end
    function [ c, ceq, gc, gceq ] = const_fish(a0)

        ceq = [
            P_shell(CX)-P_out(CX);  
            P_shell(CZ)-P_out(CZ); 
            ];
        c = [ 
            P_shell(CY)-P_out(CY);     
            ];
        
        if nargout>2
            gceq = [
                DP_shell(CX,:)-DP_out(CX,:);
                DP_shell(CZ,:)-DP_out(CZ,:);
                ]';
            gc = [ 
                DP_shell(CY,:)-DP_out(CY,:);
                ]';
        end
    end

% --------------------------------------------------------------------- 

    function [ f, gf ] = obj_standup(a0)
        
        % get displacements
        d0 = DdDa*a0;
        a1 = a0(k+1:end);
        
        % moments with respect to (d0)
        [ P1, DmDd ] = comp_moments_div(d0);
             
        % this term minimizes the volume
        f1 = P1(M1)^1;
        
        % this term minimizes mass-times-hight of the center of mass (c(2)=cy)
        % the center of mass should be low, the mass as well
        f2 = P1(CY)^1; % + P1(CX)^2 + P1(CZ)^2;
         
        % this term penalizes outer displacement
        f5 = sum(a1.^2);
        
        %objective weighted by custom weights w
        f = w(1)*f1 + w(2)*f2 +  w(6)*f5;
        %f = w(2)*(f1 * f2) + w(3)*f3 + w(4)*f4 + w(6)*f5;
        fhist = [fhist, f];
        fmap('Etot') = f;
        fmap('Ered') = f-w(6)*f5;
        
        if nargout>1
            % compute final derivatives
            ds1 = DmDd*DdDa;
            %DM_out = DmDd_out*DdDa;
            %DM_in = DmDd_in*DdDa;

            %gf1 = 2*M(M1)*DM(M1, :)';
            gf1 = ds1(M1, :)';
            gf2 = ds1(CY, :)'; % + 2*P1(CX)*ds1(CX, :)'+2*P1(CZ)*ds1(CZ, :)';
            
            gf5 = zero0;
            gf5(k+1:end) = 2*a1;  
            
            gf = w(1)*gf1 + w(2)*gf2 + w(6)*gf5;
            %gf = w(2)*(f1*gf2 + f2*gf1) + w(3)*gf3 + w(6)*gf5;
        end
    end
    function [ c, ceq, gc, gceq ] = const_standup(a0)

        ceq = [
            P1(CX);
            P1(CZ);
            ];
        c = [ 
            -P1(CY);
            P1(CY) - mesh.radii(2);
            ];
        
        if nargout>2
            % gradient
            % is computed in the objective: 
            % DM = DmDd*DdDa;
            gceq = [
                 ds1(CX,:);
                 ds1(CZ,:);
                ]';
            gc = [ 
                -ds1(CY,:);
                ds1(CY,:);
                ]';
        end
    end


% =====================================================================

    function [ M, DmDd, p ] = comp_moments_div(dio0)
           
        % get new vertices
        if(approach==1)
            % variable inner points pi            
            % fix displacement for outer points po
            dio0 = [ dio0; mesh.dfo ];        
        elseif(approach==3)           
            % fix displacement for inner points pi            
            % variable outer points po
            dio0 = [ mesh.dfi; dio0 ];                    
        end
        
        % apply spatially fixed displacements
        dio0(pio_fix) = pio_fixval; 
        % diplace points
        p = pio0 + bsxfun(@times, dio0, nio0);           
        % get points in 3-block (for each triangle)
        p = p(tri_inout,:);
        
        if nargout>1
            % get normals for inner and outer points
            % n1 = [ ni0 ; no0 ]; --> since fix, moved to preamble
            % reorder points, each row is a triangle,
            % the normals are needed for the gradient computation only here!
            [ M, ~, DmDd ] = comp_moments_grad_3d( p(1:ntri_inout,:), p(ntri_inout+1:2*ntri_inout,:), p(2*ntri_inout+1:3*ntri_inout,:), tri_inout, nio0 );
            
            if(approach==1)
                % take only the derivatives of the inner points here
                DmDd = DmDd(:,1:nn);
            elseif(approach==3)
                % take only the derivatives of the outrer points here
                DmDd = DmDd(:,nn+1:nn*2);
            end
        else
            % reorder points, each row is a triangle,
            [ M ] = comp_moments_3d( p(1:ntri_inout,:), p(ntri_inout+1:2*ntri_inout,:), p(2*ntri_inout+1:3*ntri_inout,:) );
        end        
    end

    function [ M, DmDd, p_inout,  M_out, DmDd_out, p_out,  M_in, DmDd_in, p_in ] = comp_moments_volumes_div(dio0)
        
        % get new vertices        
        if(approach==1)
            % variable inner points pi            
            % fix displacement for outer points po
            p = pio0 + bsxfun(@times, [ dio0; mesh.dfo ], nio0);            
        elseif(approach==3)           
            % fix displacement for inner points pi            
            % variable outer points po
            p = pio0 + bsxfun(@times, [ mesh.dfi; dio0 ], nio0);                        
        else
            p = pio0 + bsxfun(@times, dio0, nio0);
        end
               
        % get points in 3-block (for each triangle) for all
        p_inout = p(tri_inout,:);
        % get points in 3-block (for each triangle) for inner
        p_in    = p(1:np_in,:);        
        p_in    = p_in(tri_in,:);
        % get points in 3-block (for each triangle) for outer
        p_out   = p(np_in+1:np_in+np_out,:);        
        p_out   = p_out(tri_out,:);
                        
            % get normals for inner and outer points
            % n1 = [ ni0 ; no0 ]; --> since fix, moved to preamble
            % reorder points, each row is a triangle,
            
            % The normals are needed for the computation of the derivatives DX/Dd,
            % the sparse-normals matrix. These normals must be exactlye these which
            % are used for offseting the surface!
            
            % volume between surfaces
            [ M, ~, DmDd ] = comp_moments_grad_3d( p_inout(1:ntri_inout,:), p_inout(ntri_inout+1:2*ntri_inout,:), p_inout(2*ntri_inout+1:3*ntri_inout,:), tri_inout, nio0 );
            
            % moments of the very inner surface --- the void
            %[ M_in, ~, DmDd_in ] = comp_moments_grad_3d( p_in(1:ntri_in,:), p_in(ntri_in+1:2*ntri_in,:), p_in(2*ntri_in+1:3*ntri_in,:), tri_in, nio0(1:np,:) );
            [ M_in, ~, DmDd_in ] = comp_moments_grad_3d( p_in(1:ntri_in,:), p_in(ntri_in+1:2*ntri_in,:), p_in(2*ntri_in+1:3*ntri_in,:), tri_in, nio0(1:np,:) );
            %[ M_in ] = comp_moments_3d( p_in(1:ntri_in,:), p_in(ntri_in+1:2*ntri_in,:), p_in(2*ntri_in+1:3*ntri_in,:) );            
            
            % moments of the outer surface (compute volume only in future)
            [ M_out, ~, DmDd_out ] = comp_moments_grad_3d( p_out(1:ntri_out,:), p_out(ntri_out+1:2*ntri_out,:), p_out(2*ntri_out+1:3*ntri_out,:), tri_out, nio0(np+1:np*2,:) );
            
            
            if(approach==1)
                % take only the derivatives of the inner points here
                DmDd     = DmDd(:,1:nn);
                %DmDd_out = DmDd_out(:,1:nn);
                %DmDd_in  = DmDd_in(:,1:nn);       
            elseif(approach==3)
                % take only the derivatives of the outrer points here
                DmDd = DmDd(:,nn+1:nn*2);
                %DmDd_out = DmDd_out(:,nn+1:nn*2);
                %DmDd_in  = DmDd_in(:,nn+1:nn*2);    
            end
%         else
%             % volume between surfaces
%             [ M ] = comp_moments_3d( p_inout(1:ntri_inout,:), p_inout(ntri_inout+1:2*ntri_inout,:), p_inout(2*ntri_inout+1:3*ntri_inout,:) );
%             
%             % moments of the outer surface (compute volume only in future)
%             [ M_out ] = comp_moments_3d( p_out(1:ntri_out,:), p_out(ntri_out+1:2*ntri_out,:), p_out(2*ntri_out+1:3*ntri_out,:) );
%             
%             % moments of the very inner surface --- the void
%             [ M_in ] = comp_moments_3d( p_in(1:ntri_in,:), p_in(ntri_in+1:2*ntri_in,:), p_in(2*ntri_in+1:3*ntri_in,:) );            
%         end
    end

% ---------------------------------------------------------------------

% =====================================================================
end






